/*
**  Module    : lists.c
**  Version   : 1.0
**  Date      : 
**  Authors   : Damiano Carra
**
**  Synopsis  : this module contains the functions dealing with the
**              management of the Event List
**
**  Functions : void  insert_event(Event *)
**              void  dequeue_event(Event *)
**              void  insert_free_events(Event *);
**             Event *get_free_events(void);
**             Event *new_event (void);
**              void  init_event_list(void);
**
**  History   : 19-JAN-2006 - Creation of the module.           
**
*/
 
# ifndef        STREAMING_H
#include "streaming.h"
# endif

/***************************
**  Constant definitions  ** 
***************************/
                
# ifndef START_FLAG             /* Initial value of the number of flags */
#define  START_FLAG     4       /* used to speed up the insertion of new*/
# endif                         /* records in the lists; it must be 2^n */

# ifndef MAX_STEP               /* Maximum number of elements in a      */
#define  MAX_STEP       50      /* single sublist.                      */
# endif

                          
/********************************************
**  Global (referred to module) Variables  **
********************************************/

Event 	**event_flags;          /* Array of pointers to the access      */
                                /* points in the event list.            */
int num_event_flag;             /* Number of flags currently used in    */
                                /* the event list.                      */

/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : void insert_event(Event *event)
**  Return    : none
**  Remarks   : 'event' points to the event that must be inserted into the 
**              Event List. The Event List is a double linked circular list 
**              with an empty head. 'head_of_event_list' is the pointer
**              to the empty head that is initialized with an 'time' field
**              equal to the minimum value of a 'double'.
**              The event insertion exploits the presence of flags 
**              addressing partial sublists inside the Event List in order 
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void insert_event(Event *event)
{
 Event          *cur;
 double         sched_time;
 long           half_of_list = sched_events_cnt >> 1;
 int            flag_number = 0;

 ++sched_events_cnt;

/* The test for insertion starts from the event with the highest        */
/* 'time' field.                                                     */
 sched_time = event->time;
 cur = head_of_event_list->prev;

/* The first step towards insertion consists in the dichotomous search  */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming event is lower than the one of   */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the event is placed at the end of the Event List.          */
 if(cur->time > sched_time)  
   {                            
    flag_number = num_event_flag >> 1;  
    half_of_list = num_event_flag >> 2; 
      
    while(half_of_list)
      {
       if(event_flags[flag_number]->time > sched_time)
          flag_number += half_of_list;   
       else
         flag_number -=half_of_list;
        
       half_of_list >>= 1;
      } 

    cur = event_flags[flag_number];

/* The linear search through the sublist is started here.               */ 
    while(cur->time < sched_time)
       cur = cur->next;
    while(cur->time > sched_time)
       cur = cur->prev;
   }

/* Insertion of the Event record.                                       */
 event->prev = cur;
 event->next = cur->next;
 event->next->prev = event;
 cur->next = event;     
}

    
/*
**  Function  : void dequeue_event(Event *event)
**  Return    : none
**  Remarks   : this function removes the realized event from the circular 
**              list it belongs to by connecting its predecessor with its 
**              successor. 
**              The array of companion flags is updated every time the first 
**              sublist gets empty.
*/
void dequeue_event(Event *event)
{
 Event          *cur;
 register int   i, j;
 long           step;
 
 --sched_events_cnt;

 event->prev->next = event->next;
 event->next->prev = event->prev;

/* Each time an event is drawn from the Event List the state of the     */
/* pointers used as companion flags is checked: if at least two flags   */
/* are pointing to the outgoing event the set of flags needs to be      */
/* rearranged over the Event List.                                      */
 if((event_flags[num_event_flag - 1] == event) || 
    (event_flags[0] == event_flags[1]))
   {                                      
    step = sched_events_cnt / (num_event_flag + 1);

    if(step > MAX_STEP)
      {
       free(event_flags);
       num_event_flag <<= 1;
       event_flags = (Event **) malloc ((sizeof(Event *)) * num_event_flag);
           if ( event_flags == NULL)
             {
             printf ("Error  : Memory allocation error\n");
	         exit (1);
             }      
       step = sched_events_cnt / (num_event_flag + 1);
      }
            
    cur = head_of_event_list;
    for(j = 0; j < num_event_flag; j++)
      {
       for(i = 0; i < step; i++)
            cur = cur->prev;
       event_flags[j] = cur;
      }      
   }
 insert_free_events (event);
}




/*
**  Function  : void insert_free_events(Event *event)
**  Return    : none
**  Remarks   : instead of freeing and allocating memory for events, 
**              the program puts released event in a list of free events
**              and, when a the simulation routine needs new events, it 
**              it picks up them from this list of free events;
**              this function puts the passed event into the linera list 
**              of free events (LIFO discipline is used) 
*/
void insert_free_events (Event *elem)
{
  if (elem == NULL)
    {
      printf ("Error: event NULL\n");
      exit (1);
    }
  elem->prev = NULL;
  elem->next = free_events;
  free_events = elem;          
}


/*
**  Function  : Event *get_free_events(void)
**  Return    : pointer to an event
**  Remarks   : instead of freeing and allocating memory for events, 
**              the program puts released event in a list of free events
**              and, when a the simulation routine needs new events, it 
**              it picks up them from this list of free events;
**              this function gets free events from free_events list and 
**              makes them available to the routine that has requested them
*/
Event *get_free_events (void)
{
  Event *dummy;

  dummy = free_events;
  free_events = free_events->next; 
  return dummy;
}


/*
**  Function  : Event *new_event(void)
**  Return    : pointer to a new event
**  Remarks   : this routine returns the pointer to a new event; the event is
**              picked up from the free_events list; if no free events are 
**              available, a number of BUNCHSIZE event are allocated 
*/
Event *new_event (void)
{
  Event *dummy;
  int i;

  if (free_events == NULL)
    {
      if ((dummy = (Event *) malloc (sizeof (Event))) == NULL)
      	{
      	  printf ("Error  : Memory allocation error\n");
	      exit (1);
        }
      dummy->time = (double) 0;
      insert_free_events(dummy);
      for (i = 0; i < BUNCHSIZE; i++)
	     {
	      if ((dummy = (Event *) malloc (sizeof (Event))) == NULL)
	        {
	         printf ("Error  : Memory allocation error\n");
             exit (1);
	        }
           dummy->time = (double) 0;
	       insert_free_events(dummy);
         }/*for*/
    }/*if(free_events ==NULL)*/
  return (get_free_events());
}

/*
**  Function  : void init_event_list(void)
**  Return    : none
**  Remarks   : this function initializes the Event list allocating the head
**              element: 'head_of_event_list' and setting its 'time' field
**              to -1.0 as required by the insertion algorithm. 
**              The array of pointers supporting the Event List management 
**              is also initialized.
*/
void init_event_list(void)
{
 int i;

 if ((head_of_event_list = (Event *) malloc (sizeof (Event))) == NULL)
    {
    printf ("Error  : Memory allocation error\n");
	exit (1);
    }

 head_of_event_list->time = -1.0;
 head_of_event_list->type = HEAD;
 head_of_event_list->next = head_of_event_list;
 head_of_event_list->prev = head_of_event_list;

 num_event_flag = START_FLAG;
 if ((event_flags = (Event **) malloc ((sizeof(Event *)) * num_event_flag)) == NULL)
    {
    printf ("Error  : Memory allocation error\n");
	exit (1);
    }

 for(i = 0; i < num_event_flag; i++)
     event_flags[i] = head_of_event_list;

 /* after main event list initialization, free event list pointer is set */    
 free_events=NULL; 
}
 
/*
**  Function  : void dequeue_specific_event(int type, double time, int peer)
**  Return    : none
**  Remarks   : 'event' points to the event that must be inserted into the 
**              Event List. The Event List is a double linked circular list 
**              with an empty head. 'head_of_event_list' is the pointer
**              to the empty head that is initialized with an 'time' field
**              equal to the minimum value of a 'double'.
**              The event insertion exploits the presence of flags 
**              addressing partial sublists inside the Event List in order 
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void dequeue_specific_event(int type, double sched_time, Node *node_id)
{
 Event          *cur, *cur_ch;
 long           half_of_list = sched_events_cnt >> 1, step;
 int            flag_number = 0, rearrange, delete;
 register int   i, j;


/* The test for insertion starts from the event with the highest        */
/* 'time' field.                                                     */
  cur = head_of_event_list->prev;

/* The first step towards search consists in the dichotomous analysis   */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming event is lower than the one of   */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the event is placed at the end of the Event List.          */
 if(cur->time > sched_time)  
   {                            
    flag_number = num_event_flag >> 1;  
    half_of_list = num_event_flag >> 2; 
      
    while(half_of_list)
      {
       if(event_flags[flag_number]->time > sched_time)
          flag_number += half_of_list;   
       else
         flag_number -=half_of_list;
        
       half_of_list >>= 1;
      } 

    cur = event_flags[flag_number];

/* The linear search through the sublist is started here.               */ 
    while((cur->time <= sched_time)&&((cur->peer_id != node_id)||(cur->type != type)))
       cur = cur->next;
    while((cur->time >= sched_time)&&((cur->peer_id != node_id)||(cur->type != type)))
       cur = cur->prev;
   }

/* final check if the selected event is correct */   
  delete = FALSE;
  if ((cur->time == sched_time)&&(cur->peer_id == node_id)&&(cur->type == type))
     delete = TRUE;
  
//  if (delete == TRUE){ 
     
/* dequeue of the Event record.                                       */
  --sched_events_cnt;

 cur->prev->next = cur->next;
 cur->next->prev = cur->prev;

/* Each time an event is drawn from the Event List the state of the     */
/* pointers used as companion flags is checked: if a flag is            */
/* pointing to the outgoing event the set of flags needs to be          */
/* rearranged over the Event List.                                      */
 rearrange = FALSE;
 for(j = 0; j < num_event_flag; j++){
     if (event_flags[j] == cur)
         rearrange = TRUE;
 }
  
 if(rearrange == TRUE)
   {                                      
    step = sched_events_cnt / (num_event_flag + 1);

    if(step > MAX_STEP)
      {
       free(event_flags);
       num_event_flag <<= 1;
       event_flags = (Event **) malloc ((sizeof(Event *)) * num_event_flag);
           if ( event_flags == NULL)
             {
             printf ("Error  : Memory allocation error\n");
	         exit (1);
             }      
       step = sched_events_cnt / (num_event_flag + 1);
      }
            
    cur_ch = head_of_event_list;
    for(j = 0; j < num_event_flag; j++)
      {
       for(i = 0; i < step; i++)
            cur_ch = cur_ch->prev;
       event_flags[j] = cur_ch;
      }      
   }
 insert_free_events (cur);
// }
    
}
